from typing import Optional

from amniotic.config import MAC_ADDRESS
from amniotic.mqtt.tools import sanitize
from amniotic.version import __version__


class Device:
    """

    Representation of the parent device for Home Assistant.

    """
    NAME = 'Amniotic'
    MODEL = 'Amniotic'
    MANUFACTURER = 'Frontmatter'
    URL = None

    def __init__(self, location: Optional[str] = None):
        """

        Set any specified arguments.

        """
        self.location = location

    @property
    def uid(self) -> str:
        """

        Home Assistant compatible unique ID.

        """
        uid = sanitize(self.name, MAC_ADDRESS)
        return uid

    @property
    def name(self) -> str:
        """

        Home Assistant compatible device name.

        """

        name = self.NAME
        if self.location:
            name = f'{self.location} {name}'

        return name

    @property
    def topic_lwt(self) -> str:
        """

        Device LWT path.

        """
        subpath = sanitize(self.name, sep='/')
        topic = f'tele/{subpath}/LWT'
        return topic

    @property
    def announce_data(self) -> dict:
        """

        Home Assistant announce data for the device.

        """
        data = {
            "connections": [["mac", self.uid]],
            'sw_version': __version__,
            'name': self.name,
            'model': self.MODEL,
            'manufacturer': self.MANUFACTURER,
            'identifiers': self.uid
        }
        return data



import logging
from _socket import gethostname
from dataclasses import dataclass
from os import getenv
from pathlib import Path

import yaml
from appdirs import AppDirs
from getmac import getmac

NAME = 'amniotic'
ORG = 'frontmatter'
APP_DIRS = AppDirs(NAME, ORG)
MAC_ADDRESS = getmac.get_mac_address().replace(':', '')
HOSTNAME = gethostname()


@dataclass
class Config:
    """

    """
    mqtt_host: str = 'homeassistant.local'
    mqtt_port: int = 1883
    mqtt_username: str = None
    mqtt_password: str = None
    location: str = None
    path_audio: str = APP_DIRS.user_data_dir
    device_names: dict = None
    logging: str = None
    tele_period: int = 300

    def __post_init__(self):
        path_audio = Path(self.path_audio).absolute()
        if not path_audio.exists():
            logging.warning(f'Audio path not found: "{path_audio}"')

        self.logging = self.logging or logging.INFO

    @classmethod
    def from_file(cls):

        path_config_base = getenv('SC_CONFIG_BASE')
        if not path_config_base:
            path_config_base = APP_DIRS.user_config_dir

        path_config_base = Path(path_config_base).absolute()
        path_config_base.mkdir(parents=True, exist_ok=True)
        path_config = path_config_base / 'config.yml'

        if not path_config.exists():
            msg = f'Config file not found at "{path_config}". Default values will be used.'
            logging.warning(msg)
            config = {}
        else:
            msg = f'Config file found at "{path_config}"'
            logging.info(msg)
            config = yaml.safe_load(Path(path_config).read_text())

        config = cls(**config)
        return config



import json
import logging
import re
from dataclasses import dataclass
from time import sleep
from typing import Callable, Any

WHITESPACE = re.compile('[\s\-_]+')


@dataclass
class Message:
    """

    Object representing an MQTT message.

    """
    method: Callable
    topic: str
    data: Any = None
    serialize: bool = False
    is_announce: bool = False

    def __post_init__(self):
        """

        If data requires JSON serialization, apply it.

        """
        if self.serialize:
            self.data = json.dumps(self.data)

    @property
    def is_publish(self):
        """

        Is this a publish message?

        """
        return self.method.__name__ == 'publish' and not self.is_announce

    @property
    def is_subscribe(self):
        """

        Is this a subscribe message?

        """
        return self.method.__name__ == 'subscribe'

    def __str__(self):
        """

        String representation from logging etc.

        """
        return f'{self.method.__name__}:{self.topic}>{self.data}'

    def send(self):
        """

        Send the message by applying the method to the data.

        """

        args = [] if self.data is None else [self.data]
        self.method(self.topic, *args, qos=1)

    @classmethod
    def send_many(self, messages: list['Message'], delay: float = 0.5):
        """

        Send message types (announce/sub/pub) in distinct batches.

        """
        announces = [message for message in messages if message.is_announce]
        subscriptions = [message for message in messages if message.is_subscribe]
        publishes = [message for message in messages if message.is_publish]
        batches = [announces, subscriptions, publishes]

        do_delays = [True, True, False]

        for batch, do_delay in zip(batches, do_delays):

            for message in batch:
                logging.info(f'Queue: {message}')
                message.send()

            if batch and do_delay:
                sleep(delay)


def sanitize(*strings, sep: str = '-') -> str:
    """

    Replace spaces with URL- and ID-friendly characters, etc.

    """

    strings = [string for string in strings if string]
    string = ' '.join(strings)
    strings = [c.lower() for c in string if c.isalnum() or c in {' '}]
    string = ''.join(strings)
    string = WHITESPACE.sub(sep, string).strip()

    return string





from pathlib import Path

from setuptools import find_packages, setup

packages = find_packages()
name = next(iter(packages))
path_base = Path(__file__).absolute().parent
path = path_base / name / 'version'
__version__ = path.read_text().strip()

setup(
    long_description=(path_base / 'readme.md').read_text(),
    long_description_content_type='text/markdown',
    name=name,
    version=__version__,
    url=f'https://github.com/fmtr/{name}',
    license='Copyright © 2022 Frontmatter. All rights reserved.',
    author='Frontmatter',
    description='A multi-output ambient sound mixer for Home Assistant',
    keywords='ambient sound audio white noise masking sleep',
    packages=packages,
    package_data={
        name: [f'version'],
    },
    install_requires=[
        'paho-mqtt',
        'python-vlc',
        'getmac',
        'pyyaml',
        'appdirs',
        'johnnydep',
        'pytube'
    ],
    extras_require={},
    entry_points={
        'console_scripts': [
            'amniotic = amniotic.mqtt.loop:start',
        ],
    }
)



import json
import logging
from functools import cached_property
from json import JSONDecodeError
from time import sleep
from typing import Type

from paho.mqtt import client as mqtt

from amniotic.audio import Amniotic
from amniotic.config import Config
from amniotic.mqtt.device import Device
from amniotic.mqtt.tools import Message
from amniotic.version import __version__


class Loop:
    """

    MQTT Event Loop

    """

    CONNECTION_MESSAGES = [
        "Connection successful",
        "Connection refused - incorrect protocol version",
        "Connection refused - invalid client identifier",
        "Connection refused - server unavailable",
        "Connection refused - bad username or password",
        "Connection refused - not authorised",
    ]
    LOOP_PERIOD = 1
    DELAY = 0.5
    DELAY_FIRST = 3

    def on_message(self, client: mqtt.Client, amniotic: Amniotic, mqtt_message: mqtt.MQTTMessage):
        """

        Wrapper callback. Process payload and select and call the relevant entity object callback handler (`handle_incoming`) method.

        """

        func = self.callback_map[mqtt_message.topic]

        try:
            value = json.loads(mqtt_message.payload.decode())
        except JSONDecodeError:
            value = mqtt_message.payload.decode()

        logging.info(f'Incoming: {Message(func, mqtt_message.topic, value)}')

        return func(value)

    def on_connect_fail(self, client: mqtt.Client, amniotic: Amniotic):
        """

        Connection failed callback.

        """
        logging.error('Connection to MQTT lost.')

    def on_connect(self, client: mqtt.Client, amniotic: Amniotic, flags: dict, code: int):
        """

        Connection established/failed callback.

        """

        msg = f'Attempting to connect to MQTT "{client._host}:{client._port}": {self.CONNECTION_MESSAGES[code]}'
        if code:
            logging.error(msg)
        else:
            logging.info(msg)

        self.has_reconnected = True

    def __init__(self, host, port, device: Device, amniotic: Amniotic, username: str = None, password: str = None,
                 tele_period: int = 300):
        """

        Setup and connect MQTT Client.

        """

        self.device = device

        self.exit_reason = False

        self.entities = {
            entity_class: entity_class(self)
            for entity_class in self.entity_classes
        }
        self.callback_map = {
            entity.topic_command: entity.handle_incoming
            for entity in self.entities.values()
        }

        self.queue = []
        self.tele_period = tele_period
        self.has_reconnected = True
        self.topic_lwt = self.device.topic_lwt

        self.amniotic = amniotic
        self.client = mqtt.Client()

        self.client.on_connect = self.on_connect
        self.client.on_message = self.on_message
        self.client.on_connect_fail = self.on_connect_fail
        self.client.user_data_set(amniotic)
        self.client.will_set(self.topic_lwt, payload='Offline', qos=1, retain=False, properties=None)

        if username is not None and password is not None:
            self.client.username_pw_set(username=username, password=password)

        self.client.connect(host=host, port=port)

    @cached_property
    def entity_classes(self) -> list[Type]:
        """

        Import all entity/sensor classes

        """
        from amniotic.mqtt import control, sensor

        controls = [
            control.SelectTheme,
            control.VolumeMaster,
            control.VolumeTheme,
            control.ToggleTheme,
            control.DeviceTheme,
            control.ButtonUpdateCheck,
            control.ButtonUpdate,
            control.Downloader,
            control.NewTheme,
        ]

        sensors = [
            sensor.Title,
            sensor.Album,
            sensor.TrackCount,
            sensor.Date,
            sensor.By,
            sensor.Duration,
            sensor.Elapsed,
            sensor.UpdateStatus,
            sensor.DownloaderStatus,
            sensor.Version
        ]

        return controls + sensors

    def handle_outgoing(self, force_announce=False):
        """

        Call entity outgoing methods to add to message queue.

        """

        for entity in self.entities.values():
            entity.handle_outgoing(force_announce=force_announce)

    def do_telemetry(self):
        """

        Send LWT message.

        """
        status = json.dumps(self.amniotic.status)
        logging.info(f'Telemetry: LWT')
        logging.debug(f'Status: {status}')
        # self.client.publish(TOPIC_STATUS, status)
        self.client.publish(self.topic_lwt, "Online", qos=1)
        self

    def loop_start(self):
        """

        Run Event Loop. Once connected, periodically aggregate entity messages into the queue, send queue messages, send LWT/telemetry.

        """

        self.client.loop_start()

        loop_count = 0

        while not self.client.is_connected():
            sleep(self.LOOP_PERIOD)

        while not self.exit_reason:

            if not loop_count:
                delay = self.DELAY_FIRST
            else:
                delay = self.DELAY

            if not self.client.is_connected():
                sleep(self.LOOP_PERIOD)
                continue

            is_telem_loop = loop_count % self.tele_period == 0

            self.handle_outgoing(force_announce=self.has_reconnected)
            self.has_reconnected = False

            Message.send_many(self.queue, delay=delay)
            self.queue.clear()

            if is_telem_loop:
                self.do_telemetry()

            sleep(self.LOOP_PERIOD)
            loop_count += 1

        msg = f'Event loop exiting gracefully for: {self.exit_reason}'
        logging.info(msg)


def start():
    """

    Load config, set up amniotic, MQTT devices and entities, and start MQTT event loop.

    """
    config = Config.from_file()
    logging.basicConfig(
        format='%(asctime)s %(levelname)-5s amniotic.%(module)-8s: %(message)s',
        level=config.logging,
        datefmt='%Y-%m-%d %H:%M:%S',
        force=True
    )

    amniotic = Amniotic(path=config.path_audio, device_names=config.device_names)
    msg = f'Amniotic {__version__} has started.'
    logging.info(msg)
    msg = f'Amniotic {__version__} starting MQTT...'
    logging.info(msg)

    loop = Loop(
        device=Device(location=config.location),
        amniotic=amniotic,
        host=config.mqtt_host,
        port=config.mqtt_port,
        username=config.mqtt_username,
        password=config.mqtt_password,
        tele_period=config.tele_period
    )

    loop.loop_start()


if __name__ == '__main__':
    start()



import logging
import threading
from functools import cached_property
from time import sleep
from typing import Optional, Any

import pip
from johnnydep import JohnnyDist as Package
from paho.mqtt import client as mqtt
from pytube import YouTube, Stream

from amniotic.audio import Amniotic
from amniotic.config import NAME
from amniotic.mqtt.device import Device
from amniotic.mqtt.loop import Loop
from amniotic.mqtt.tools import Message, sanitize
from amniotic.version import __version__


class Entity:
    """

    Base representation of an entity for Home Assistant.

    """
    PAYLOAD_ONLINE = "Online"
    PAYLOAD_OFFLINE = "Offline"
    HA_PLATFORM = None
    NAME = None
    ICON_SUFFIX = None
    value = None

    def __init__(self, loop: Loop):
        self.loop = loop

    @property
    def name(self) -> str:
        """

        Home Assistant compatible entity name.

        """

        name = f'{self.device.name} {self.NAME}'
        return name

    @property
    def device(self) -> Device:
        return self.loop.device

    @property
    def queue(self) -> list[Message]:
        return self.loop.queue

    @property
    def client(self) -> mqtt.Client:
        return self.loop.client

    @property
    def amniotic(self) -> Amniotic:
        return self.loop.amniotic

    @property
    def uid(self) -> str:
        """

        Unique ID

        """
        return sanitize(self.name, sep='_')

    @property
    def topic_state(self) -> str:
        """

        State topic

        """
        subpath = sanitize(self.name, sep='/')
        topic = f'stat/{subpath}/state'
        return topic

    @property
    def topic_command(self) -> str:
        """

        Command topic

        """
        subpath = sanitize(self.name, sep='/')
        topic = f'stat/{subpath}/command'
        return topic

    @property
    def topic_announce(self) -> str:
        """

        Topic to announce to Home Assistant

        """
        subpath = sanitize(self.name, sep='_')
        topic = f'homeassistant/{self.HA_PLATFORM}/{subpath}/config'
        return topic

    @property
    def icon(self) -> Optional[str]:
        """

        Add Material Design Icons prefix to icon name.

        """
        if not self.ICON_SUFFIX:
            return self.ICON_SUFFIX

        icon = f"mdi:{self.ICON_SUFFIX}"
        return icon

    @property
    def data(self) -> dict:
        """

        Home Assistant announce data for the entity.

        """
        data = {
            "name": self.name,
            "unique_id": self.uid,
            "object_id": self.uid,
            "device": self.device.announce_data,
            "device_class": self.HA_PLATFORM,
            "force_update": True,
            "payload_available": self.PAYLOAD_ONLINE,
            "payload_not_available": self.PAYLOAD_OFFLINE,
            "availability_topic": self.device.topic_lwt,
            "state_topic": self.topic_state,
            "command_topic": self.topic_command,
        }

        if self.icon:
            data['icon'] = self.icon

        return data

    def get_value(self) -> Any:
        raise NotImplementedError()

    def set_value(self, value) -> Any:
        raise NotImplementedError()

    def handle_incoming(self, value: Any):
        """

        Apply change audio volume from incoming message.

        """
        if value is not None:
            self.set_value(value)

    def handle_announce(self):
        message = Message(self.client.publish, self.topic_announce, self.data, serialize=True, is_announce=True)
        self.queue.append(message)
        if self.topic_command:
            message = Message(self.client.subscribe, self.topic_command)
            self.queue.append(message)

    def handle_outgoing(self, force_announce: bool = False):
        """

        Handle outgoing messages, adding announce, subscriptions to the queue.

        """
        if force_announce:
            self.handle_announce()

        value = self.get_value()
        if value != self.value or force_announce:
            self.set_value(value)
            self.value = self.get_value()
            message = Message(self.client.publish, self.topic_state, self.value)
            self.queue.append(message)


class Select(Entity):
    """

    Base Home Assistant selector.

    """
    HA_PLATFORM = 'select'
    ICON_SUFFIX = None
    options: list[str] = []
    selected: Optional[str] = None

    @property
    def data(self):
        """

        Home Assistant announce data for the entity.

        """
        data = super().data | {
            'options': self.options
        }
        return data

    def get_options(self, amniotic: Amniotic) -> list[str]:
        """

        Get state of the entity, i.e. the list of options and the currently selected option.

        """
        raise NotImplementedError()

    def handle_outgoing(self, force_announce: bool = False):
        """

        Handle outgoing messages, adding announce, subscriptions to the queue.

        """

        options = self.get_options(self.amniotic)
        if options != self.options or force_announce:
            force_announce = True
            self.options = options

        super().handle_outgoing(force_announce=force_announce)


class SelectTheme(Select):
    """

    Home Assistant theme selector.

    """
    ICON_SUFFIX = 'surround-sound'
    NAME = 'Theme'

    def get_value(self) -> Any:
        return self.amniotic.theme_current.name

    def set_value(self, value) -> Any:
        self.amniotic.set_theme(value)

    def get_options(self, amniotic: Amniotic) -> list[str]:
        """

        Get state of the entity, i.e. the list of options and the currently selected option.

        """
        return list(amniotic.themes.keys())


class Volume(Entity):
    """

    Home Assistant base volume control.

    """
    HA_PLATFORM = 'number'
    MIN = 0
    MAX = 100

    @property
    def data(self):
        data = super().data | {
            'min': self.MIN,
            'max': self.MAX
        }
        return data


class VolumeMaster(Volume):
    """

    Home Assistant master volume control.

    """
    HA_PLATFORM = 'number'
    ICON_SUFFIX = 'volume-high'
    NAME = 'Master Volume'

    @property
    def data(self):
        """

        Home Assistant announce data for the entity.

        """
        data = super().data | {
            'min': self.MIN,
            'max': self.MAX
        }
        return data

    def get_value(self) -> Any:
        return self.amniotic.volume

    def set_value(self, value) -> Any:
        self.amniotic.set_volume(value)


class VolumeTheme(Volume):
    """

    Home Assistant theme volume control.

    """

    ICON_SUFFIX = 'volume-medium'
    NAME = 'Theme Volume'

    def get_value(self) -> Any:
        return self.amniotic.theme_current.volume

    def set_value(self, value) -> Any:
        self.amniotic.set_volume_theme(value)


class DeviceTheme(Select):
    """

    Home Assistant device selector.

    """
    ICON_SUFFIX = 'expansion-card-variant'
    NAME = 'Theme Device'

    def get_value(self) -> Any:
        return self.amniotic.theme_current.device_name

    def set_value(self, value) -> Any:
        self.amniotic.theme_current.set_device(value)

    def get_options(self, amniotic: Amniotic) -> list[str]:
        """

        Get state of the entity, i.e. the list of options and the currently selected option.

        """
        return list(amniotic.devices.values())


class ToggleTheme(Entity):
    """

    Base Home Assistant Theme enabled/disabled entity (switch/toggle).

    """
    HA_PLATFORM = 'switch'
    ICON_SUFFIX = 'play-circle'
    VALUE_MAP = [(OFF := 'OFF'), (ON := 'ON')]
    NAME = 'Theme Enabled'

    def get_value(self) -> Any:
        return self.VALUE_MAP[self.amniotic.theme_current.enabled]

    def set_value(self, value) -> Any:
        self.amniotic.theme_current.enabled = value == self.ON

    @property
    def data(self):
        """

        Home Assistant announce data for the entity.

        """
        data = super().data | {
            'device_class': 'outlet',
        }
        return data


class ButtonUpdateCheck(Entity):
    """

    Home Assistant update button.

    """
    HA_PLATFORM = 'button'
    ICON_SUFFIX = 'source-branch-sync'
    NAME = 'Update Check'

    def get_value(self) -> Any:
        pass

    def set_value(self, value) -> Any:
        pass

    @property
    def data(self):
        """

        Home Assistant announce data for the entity.

        """
        data = super().data
        data.pop('device_class')
        return data

    @cached_property
    def update_sensor(self):
        """

        Get the sensor for displaying update messages

        """
        from amniotic.mqtt.sensor import UpdateStatus
        update_status = self.loop.entities[UpdateStatus]
        return update_status

    def get_pypi_latest(self) -> Optional[str]:
        """

        Check if newer version is available.

        """
        package = Package(NAME)
        version = package.version_latest
        if __version__ == package.version_latest:
            return None
        else:
            return version

    def check_update(self):
        """

        Report newer version, if one exists, using update sensor

        """

        try:
            version = self.get_pypi_latest()
        except Exception as exception:
            self.update_sensor.message = f'Error checking for updates ({exception.__class__.__name__})'
            return

        if version:
            message = f'Update available: {version}'
        else:
            message = 'None available'
        self.update_sensor.message = message

    def handle_incoming(self, value: Any):
        """

        When button is pressed, call update method without blocking.

        """

        self.update_sensor.message = 'Checking for updates...'
        threading.Thread(target=self.check_update).start()


class ButtonUpdate(ButtonUpdateCheck):
    """

    Home Assistant update button.

    """

    HA_PLATFORM = 'button'
    NAME = 'Update'
    ICON_SUFFIX = None

    @property
    def data(self):
        """

        Home Assistant announce data for the entity.

        """

        data = super().data | {
            'device_class': 'update'
        }

        return data

    def do_update(self):
        """

        Update from PyPI, then tell loop to exit.

        """

        try:
            pip.main(['install', NAME, '--upgrade'])
        except:
            self.update_sensor.message = 'Error updating'
            return
        self.update_sensor.message = 'Update complete. Restarting...'
        sleep(self.loop.DELAY_FIRST)
        self.loop.exit_reason = f'Updating to latest version.'

    def handle_incoming(self, value: Any):
        """

        When button is pressed, call update method without blocking.

        """

        self.update_sensor.message = 'Updating...'
        threading.Thread(target=self.do_update).start()


class TextInput(Entity):
    """

    Base Home Assistant text input box. Note: this control abuses an alarm panel code entry box, as it seems to be the only way to allow a user to send
    arbitrary text (e.g. a URL) from a Home Assistant control.

    """
    HA_PLATFORM = 'alarm_control_panel'
    DISARMED = 'disarmed'
    TRIGGERED = 'triggered'
    status = DISARMED

    @cached_property
    def update_sensor(self):
        """

        Get the sensor for displaying update messages

        """
        raise NotImplementedError()

    def set_value(self, value) -> Any:
        """

        Dummy method

        """
        pass

    def get_value(self) -> Any:
        """

        The current state of this control. Pending means Idle, and Triggered means Downloading.

        """
        return self.status

    @property
    def data(self):
        """

        Home Assistant announce data for the entity.

        """
        data = super().data | {
            'code': 'REMOTE_CODE_TEXT',
            'command_template': "{{ code }}"
        }
        return data


class NewTheme(TextInput):
    """

    Home Assistant text input box for creating a new Theme

    """
    HA_PLATFORM = 'alarm_control_panel'
    ICON_SUFFIX = 'folder-plus-outline'
    NAME = 'Create New Theme'

    def handle_incoming(self, value: Any):
        """

        Add specified theme and set to current

        """
        self.amniotic.add_new_theme(value, set_current=True)


class Downloader(TextInput):
    """

    Home Assistant track downloader URL input.

    """
    HA_PLATFORM = 'alarm_control_panel'
    ICON_SUFFIX = 'cloud-download-outline'
    NAME = 'Download YouTube Link'

    IDLE = TextInput.DISARMED
    DOWNLOADING = TextInput.TRIGGERED
    status = IDLE

    @cached_property
    def update_sensor(self):
        """

        Get the sensor for displaying update messages

        """
        from amniotic.mqtt.sensor import DownloaderStatus
        update_status = self.loop.entities[DownloaderStatus]
        return update_status

    def progress_callback(self, stream: Stream, chunk: bytes, bytes_remaining: int):
        """

        Send download progress to sensor

        """

        percentage = (1 - (bytes_remaining / stream.filesize)) * 100
        self.update_sensor.message = f'Downloading: {round(percentage)}% complete'

    def completed_callback(self, stream: Stream, path: str):
        """

        Send download completion message to sensor

        """

        self.update_sensor.message = f'Download complete: "{stream.title}"'
        self.status = self.IDLE

    def do_download(self, url: str):
        """

        Download highest bitrate audio stream from the video specified. Log any errors/progress to the relevant sensor

        """

        try:

            self.status = self.DOWNLOADING
            theme = self.amniotic.theme_current
            self.update_sensor.message = 'Fetching video metadata...'

            video = YouTube(
                url,
                on_progress_callback=self.progress_callback,
                on_complete_callback=self.completed_callback
            )
            self.update_sensor.message = 'Finding audio streams...'
            audio_streams = video.streams.filter(only_audio=True).order_by('bitrate')
            if not audio_streams:
                self.update_sensor.message = f'Error downloading: no audio streams found in "{video.title}"'
                self.status = self.IDLE
                return
            stream = audio_streams.last()

            if stream.filesize == 0:
                self.update_sensor.message = f'Error downloading: empty audio stream found in "{video.title}"'
                self.status = self.IDLE
                return

            self.update_sensor.message = 'Starting download...'
            stream.download(output_path=str(theme.path))

        except Exception as exception:

            self.update_sensor.message = f'Error downloading: {exception.__class__.__name__}'
            logging.error(f'Download error for "{url}": {repr(exception)}')
            self.status = self.IDLE
            return

    def handle_incoming(self, value: Any):
        """

        Start download from the specified URL without blocking.

        """

        if self.status == self.DOWNLOADING:
            return
        threading.Thread(target=self.do_download, args=[value]).start()





import getpass
import logging
from datetime import datetime
from itertools import cycle
from pathlib import Path
from random import choice
from typing import Union, Optional

import vlc

VLC_VERBOSITY = 0


class Amniotic:
    VOLUME_DEFAULT = 50
    THEME_NAME_DEFAULT = 'Default Theme'

    def __init__(self, path: Union[Path, str], device_names: Optional[dict[str, str]] = None):
        """

        Read audio directories and instantiate Theme objects

        """

        user = getpass.getuser()
        if user == 'root':
            msg = f'You are running as root. This could cause issues with PulseAudio, unless it is configured in system-wide mode.'
            logging.warning(msg)

        self.device_names = device_names or {}
        self._enabled = True
        path = Path(path).absolute()
        paths_themes = sorted([path.absolute() for path in path.glob('*') if path.is_dir()])

        if not paths_themes:
            msg = f'No audio directories found in "{path}". Default theme will be created.'
            logging.warning(msg)

        self.path = path

        self.themes = [Theme(path, device_names=self.device_names) for path in paths_themes]
        self.themes = {theme.name: theme for theme in self.themes}
        if not self.themes:
            self.add_new_theme(self.THEME_NAME_DEFAULT)

        self.theme_current = None
        self.set_theme(next(iter(self.themes.keys())))

        self.volume = 0
        self.set_volume(self.VOLUME_DEFAULT)

    @property
    def devices(self) -> dict[str, str]:
        """

        Get general (system-wide) devices. Just use those of the first Theme.

        """
        return self.theme_current.devices

    @property
    def enabled(self) -> bool:
        """

        Are any Themes enabled?

        """
        return any([theme.enabled for theme in self.themes.values()])

    @enabled.setter
    def enabled(self, value: bool):
        """

        If set to `False`, disable all Themes.

        """
        value = bool(value)
        self._enabled = value

        if not self._enabled:
            for theme in self.themes.values():
                theme.enabled = self._enabled

    def set_theme(self, id: str):
        """

        Set current theme by the specified name/ID.

        """
        if id not in self.themes:
            id = self.theme_current.get_device_id(id)
        self.theme_current = self.themes[id]

    def add_new_theme(self, name: str, set_current: bool = False):
        """

        Add a new, empty theme by the specified name/ID.

        """
        if name not in self.themes:
            path = self.path / name
            path.mkdir()
            theme = Theme(path, device_names=self.device_names)
            self.themes[name] = theme
            if set_current:
                self.set_theme(name)

    def set_volume(self, value: int):
        """

        Set Master Volume, and propagate to all Themes.

        """
        self.volume = value
        for theme in self.themes.values():
            theme.set_volume(self.volume)

    def set_volume_theme(self, value):
        """

        Set the current Theme Volume.

        """
        self.theme_current.set_volume(self.volume, value)

    @property
    def status(self) -> dict:
        """

        General status information, including that of all themes.

        """
        themes = [theme.status for theme in self.themes.values()]
        data = {'datetime': datetime.now().isoformat(), 'volume': self.volume, 'themes': themes}
        return data


class Theme:
    VOLUME_DEFAULT = 40

    def __init__(self, path: Path, device_names: Optional[dict[str, str]] = None):
        """

        Fetch paths from Theme audio directory, set up two alternating players (so they can call each other without thread deadlocks) and set a default
        audio output device.

        """
        self.path = path
        self.name = path.stem
        self.paths = self.get_paths()

        if not self.paths:
            msg = f'Theme "{self.name}" directory is empty: "{self.path}"'
            logging.warning(msg)

        self.device_names = device_names or {}
        self._enabled = False
        self.ever_started = False

        self.players = cycle([self.get_player(), self.get_player()])
        self.player = self.device = None
        self.switch_player()
        self.set_device(device=None)
        self.volume = self.VOLUME_DEFAULT
        self.volume_scaled = self.volume

    def update_paths(self):
        """

        Update file paths from disk.

        """
        self.paths = self.get_paths()

    def get_paths(self) -> list[Path]:
        """

        Get file paths from disk.

        """
        paths = list(self.path.glob('*'))

        return paths

    def get_player(self) -> vlc.MediaPlayer:
        """

        Instantiate a new player, and register callbacks

        """
        instance = vlc.Instance(f'--verbose {VLC_VERBOSITY}')
        player = vlc.MediaPlayer(instance)
        player.event_manager().event_attach(vlc.EventType.MediaPlayerEndReached, self.cb_media_player_end_reached)
        return player

    def cb_media_player_end_reached(self, event: vlc.Event):
        """

        Method to register as a VLC callback. Once a file finishes playing, start playing the next file in the other player.

        """
        logging.debug(f'Theme "{self.name}" hit media end callback with event: {event.type=} {event.obj=} {event.meta_type=}')
        self.switch_player()
        self.play()

    @property
    def instance(self) -> vlc.Instance:
        """

        Get the `Instance` behind the current `MediaPlayer`.

        """
        return self.player.get_instance()

    @property
    def device_name(self):
        """

        Get the (friendly) name of the current device. If it doesn't exist, it must have been unplugged etc. so the default needs setting.

        """

        if self.device not in self.devices:
            self.set_device(self.device)

        return self.devices.get(self.device)

    @property
    def devices(self) -> dict[str, str]:
        """

        Create a mapping from audio output device IDs to their friendly names from VLC's peculiar enum format.

        """

        devices_raw = self.player.audio_output_device_enum()
        devices = {}
        if devices_raw:
            device_raw = devices_raw
            count = 0
            while device_raw:
                count += 1
                device_raw = device_raw.contents
                description = device_raw.description.decode()
                device = device_raw.device.decode()
                count = len([key for key in devices.keys() if key == description])
                if count:
                    description = f'{description} ({count + 1})'
                devices[device] = self.device_names.get(description, description)
                device_raw = device_raw.next

        return devices

    def set_device(self, device: Optional[str]):
        """

        Set the output audio device from its ID. Also handle when that device had been unplugged, etc.

        """
        devices = self.devices

        if device not in devices:
            device = self.get_device_id(device)

        if device not in devices:
            self.enabled = False
            device = next(iter(devices or {None}))
            msg = f'Current device "{self.device}" no longer available for theme "{self.name}". ' \
                  f'Defaulting to "{device}". Theme will be disabled.'
            logging.warning(msg)

        self.device = device

        if self.enabled:
            self.player.audio_output_device_set(None, device)

    def get_device_id(self, name: str) -> Optional[str]:
        """

        Get a device ID from its friendly name.

        """
        device_id = {name: id for id, name in self.devices.items()}.get(name)
        return device_id

    def switch_player(self):
        """

        Alternate the current player.

        """
        self.player = next(self.players)
        logging.debug(f'Theme "{self.name}" switched player to: {self.player}')

    def play(self):
        """

        Play a single audio file at random from this Theme's directory. Many settings (e.g. the output device) will default between plays,
        so this all needs specifying each time.

        """
        self.ever_started = True
        path = choice(self.paths)
        media = self.instance.media_new(str(path))
        self.player.set_media(media)
        self.set_device(self.device)
        self.player.audio_set_volume(self.volume_scaled)
        logging.info(f'Theme "{self.name}" playing file {path}')
        self.player.play()

    @property
    def enabled(self) -> bool:
        """

        Is this theme enabled?

        """
        return self._enabled

    @enabled.setter
    def enabled(self, value: bool):
        """

        Set whether Theme is enabled. If the input value if different from current, either start playing or toggle pause, depending on Theme state. Themes
        with no tracks (paths) cannot be enabled.

        """

        if not self.paths:
            return

        value = bool(value)
        if value == self._enabled:
            return

        self._enabled = value

        if not self.ever_started and self._enabled:
            self.play()
        else:
            self.player.pause()

    def set_volume(self, volume_master: int, volume: Optional[int] = None):
        """

        Set the scaled volume by multiplying the master volume with the theme volume.

        """

        if volume is not None:
            self.volume = volume

        volume_old = self.volume_scaled
        volume_scaled = round(self.volume * (volume_master / 100))
        logging.info(f'Changing scaled volume for theme "{self.name}": from {volume_old} to {volume_scaled}')
        self.volume_scaled = volume_scaled
        self.player.audio_set_volume(volume_scaled)

    @property
    def status(self):
        """

        General Theme status information

        """
        media = self.player.get_media()

        position = self.player.get_position()
        duration = media.get_duration() if media else None

        if position and duration:
            elapsed = round(position * duration)
        else:
            elapsed = None

        data = {
            'name': self.name,
            'device': {'id': self.device, 'name': self.devices[self.device]},
            'enabled': self.enabled,
            'track_count': len(self.paths),
            'volume': {'theme': self.volume, 'scaled': self.volume_scaled},
            'position': position,
            'position_percentage': round(position * 100) if position else None,
            'elapsed': elapsed,
            'state': str(self.player.get_state()),
            'duration': media.get_duration() if media else None,
            'meta_data': {
                value: datum
                for key, value in vlc.Meta._enum_names_.items()
                if (datum := media.get_meta(key))
            } if media else None
        }

        return data



from datetime import timedelta
from typing import Optional, Union, Any

from amniotic.mqtt import control
from amniotic.version import __version__


class Sensor(control.Entity):
    """

    Base Home Assistant Theme sensor entity, sends messages taken from current Theme status data.

    """
    HA_PLATFORM = 'sensor'
    NA_VALUE = '-'
    META_KEY = None
    IS_SOURCE_META = True
    UOM = None

    @property
    def topic_command(self):
        return None

    @property
    def data(self):
        data = super().data
        data.pop('device_class')
        if self.UOM:
            data['unit_of_measurement'] = self.UOM
        return data

    def set_value(self, value: Any):
        """

        Dummy method.

        """
        pass

    def get_value(self, key=None) -> Union[str, int, float]:
        """

        Get the relevant value from the Theme status or metadata dictionaries

        """
        key = key or self.META_KEY
        status = self.amniotic.theme_current.status
        if self.IS_SOURCE_META:
            status = status.get('meta_data') or {}
        meta_value = status.get(key)
        if meta_value is None:
            meta_value = self.NA_VALUE
        return meta_value


class Title(Sensor):
    """

    Home Assistant Title sensor

    """

    NAME = META_KEY = 'Title'
    ICON_SUFFIX = 'text-recognition'


class Album(Sensor):
    """

    Home Assistant Album sensor

    """
    NAME = META_KEY = 'Album'
    ICON_SUFFIX = 'album'


class Date(Sensor):
    """

    Home Assistant Date sensor

    """
    NAME = META_KEY = 'Date'
    ICON_SUFFIX = 'calendar-outline'


class By(Sensor):
    """

    Home Assistant By (Artist) sensor

    """
    META_KEY = 'Artist'
    NAME = 'By'
    ICON_SUFFIX = 'account'


class Duration(Sensor):
    """

    Home Assistant Duration sensor

    """
    META_KEY = 'duration'
    NAME = 'Duration'
    IS_SOURCE_META = False
    ICON_SUFFIX = 'timer'

    def get_value(self, key: Optional[str] = None):
        """

        Get the value in milliseconds from the status, change to per-second granularity and return as string.

        """
        milliseconds = super().get_value()

        if milliseconds == self.NA_VALUE or milliseconds < 0:
            return super().NA_VALUE

        delta = timedelta(milliseconds=milliseconds)
        delta -= timedelta(microseconds=delta.microseconds)
        delta_str = str(delta)

        return delta_str


class Elapsed(Duration):
    """

    Home Assistant Elapsed sensor

    """
    META_KEY = 'elapsed'
    NAME = 'Elapsed'
    ICON_SUFFIX = 'clock-time-twelve-outline'


class StaticMessageSensor(Sensor):
    """

    Base Home Assistant static message sensor. Just reports a static message. The message will be set by other controls.

    """

    message = Sensor.NA_VALUE

    def get_value(self, key=None) -> Union[str, int, float]:
        """

        Return current message

        """
        return self.message


class UpdateStatus(StaticMessageSensor):
    """

    Home Assistant update status sensor. Messages set by Check Update button etc.

    """
    NAME = 'Update Status'
    ICON_SUFFIX = 'semantic-web'
    message = 'Never checked'


class DownloaderStatus(StaticMessageSensor):
    """

    Home Assistant downloader status sensor. Messages set by downloader input etc.

    """
    NAME = 'Downloader Status'
    ICON_SUFFIX = 'cloud-sync-outline'
    message = 'Idle'


class TrackCount(Sensor):
    """

    Home Assistant Track Count sensor

    """
    META_KEY = 'track_count'
    NAME = 'Track Count'
    IS_SOURCE_META = False
    ICON_SUFFIX = 'pound-box'

    def get_value(self, key=None) -> Union[str, int, float]:
        """

        Update theme from disk before reporting track count.

        """
        self.amniotic.theme_current.update_paths()
        return super().get_value(key)


class Version(StaticMessageSensor):
    """

    Home Assistant version number sensor

    """

    NAME = 'Current Version'
    ICON_SUFFIX = 'counter'
    message = __version__
